use core::error::Error;
use std::{
    fs::File,
    io::{self, Read},
    path::{Path, PathBuf},
};

use proc_macro2::Span;
use quote::{ToTokens, quote};
use syn::{parse::Parse, parse_macro_input};

struct IncludePath {
    dir_path: PathBuf,
    modules: Vec<syn::Ident>,
}

fn display_err(span: Span, err: impl Error) -> syn::Error {
    syn::Error::new(span, err.to_string().as_str())
}

fn read_modules(span: Span, path: &Path) -> syn::Result<Vec<syn::Ident>> {
    let read_dir = std::fs::read_dir(path).map_err(|err| display_err(span, err))?;

    Ok(read_dir
        .map(|file| file.map_err(|err| display_err(span, err)))
        .collect::<Result<Vec<_>, _>>()?
        .into_iter()
        .filter(|file| file.file_type().map(|ft| ft.is_file()).unwrap_or_default())
        .filter_map(|file| {
            file.file_name()
                .to_string_lossy()
                .strip_suffix(".rs")
                .map(|f| f.to_string())
        })
        .map(|module_name| syn::Ident::new(&module_name, span))
        .collect::<Vec<_>>())
}

impl Parse for IncludePath {
    fn parse(input: syn::parse::ParseStream) -> syn::Result<Self> {
        let directory = input.parse::<syn::LitStr>()?;

        let dir_path = std::env::current_dir()
            .unwrap()
            .join(directory.value())
            .to_path_buf();

        let modules = read_modules(directory.span(), &dir_path)?;

        Ok(Self { modules, dir_path })
    }
}
impl IncludePath {
    fn posts_tokens(&self) -> proc_macro2::TokenStream {
        let mod_decl = self.modules.iter().map(|module| {
            let path = self.dir_path.join(format!("{module}.rs"));
            let path = path.to_str().unwrap();
            quote! {
            mod #module {
                include!(#path);
            }}
        });
        let posts = self.modules.iter().map(|module| {
            quote! {
                #module::POST
            }
        });

        quote! {
            #(#mod_decl)*

            pub const POSTS: &[Post] = &[#(#posts),*];
        }
    }
    fn normal_tokens(&self) -> proc_macro2::TokenStream {
        let mod_decl = self.modules.iter().map(|module| {
            quote! {mod #module;}
        });
        let pub_use = self.modules.iter().map(|module| {
            quote! {
                #module::*
            }
        });

        quote! {
            #(#mod_decl)*
            pub use {#(#pub_use),*};
        }
    }
}
enum IncludeOutput {
    Normal(IncludePath),
    Posts(IncludePath),
}

impl ToTokens for IncludeOutput {
    fn to_tokens(&self, tokens: &mut proc_macro2::TokenStream) {
        tokens.extend(match self {
            IncludeOutput::Normal(include_path) => include_path.normal_tokens(),
            IncludeOutput::Posts(include_path) => include_path.posts_tokens(),
        });
    }
}

#[proc_macro]
pub fn posts(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
    let pubuse = IncludeOutput::Posts(parse_macro_input!(input as IncludePath));
    quote! {#pubuse}.into()
}

#[proc_macro]
pub fn include_path(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
    let incl_path = IncludeOutput::Normal(parse_macro_input!(input as IncludePath));
    quote! {#incl_path}.into()
}

struct FileWithPath {
    // bytes: Vec<u8>,
    include: proc_macro2::TokenStream,
    rel_path: String,
}
impl FileWithPath {
    fn from_path(path: impl AsRef<Path>, origin_path: impl AsRef<Path>) -> Result<Self, io::Error> {
        let abs_path = path
            .as_ref()
            .canonicalize()
            .ok()
            .and_then(|p| p.to_str().map(|t| t.to_string()))
            .ok_or(io::Error::new(io::ErrorKind::InvalidData, "invalid path"))?;
        let include = quote! {
            include_bytes!(#abs_path)
        };
        let origin_abs = origin_path.as_ref().canonicalize().and_then(|p| {
            p.to_str()
                .ok_or(io::Error::new(io::ErrorKind::InvalidData, "invalid path"))
                .map(|t| t.to_string())
        })?;
        let rel_path = abs_path.replace(&origin_abs, "");

        Ok(Self { include, rel_path })
    }
}

fn find_matching_files_recursive<P, F, D>(
    p: P,
    include_in_rerun: F,
    directory_filter: D,
    out: &mut Vec<FileWithPath>,
    origin_path: &PathBuf,
) -> Result<(), io::Error>
where
    P: AsRef<Path>,
    F: Fn(&str) -> bool + Copy,
    D: Fn(&str) -> bool + Copy,
{
    for item in std::fs::read_dir(p.as_ref())? {
        let item = item?;
        if item.file_type()?.is_dir() {
            if directory_filter(item.file_name().to_str().ok_or(io::Error::new(
                io::ErrorKind::InvalidData,
                "file has no name",
            ))?) {
                find_matching_files_recursive(
                    item.path(),
                    include_in_rerun,
                    directory_filter,
                    out,
                    origin_path,
                )?;
            }
            continue;
        }
        if let Some(file_name) = item.file_name().to_str() {
            if include_in_rerun(file_name) {
                out.push(FileWithPath::from_path(item.path(), origin_path)?);
            }
        }
    }

    Ok(())
}

struct IncludeDist {
    name: syn::Ident,
    files: Vec<FileWithPath>,
}

impl IncludeDist {
    fn read_dist_path(
        name: syn::Ident,
        path_span: Span,
        path: &Path,
        origin_path: &PathBuf,
    ) -> syn::Result<Self> {
        let mut files = Vec::new();
        find_matching_files_recursive(path, |_| true, |_| true, &mut files, origin_path)
            .map_err(|err| syn::Error::new(path_span, err.to_string()))?;
        Ok(Self { name, files })
    }
}

impl Parse for IncludeDist {
    fn parse(input: syn::parse::ParseStream) -> syn::Result<Self> {
        let name = input.parse::<syn::Ident>()?;
        input.parse::<syn::Token![,]>()?;
        let directory = input.parse::<syn::LitStr>()?;

        let dir_path = std::env::current_dir()
            .unwrap()
            .join(directory.value())
            .to_path_buf();

        Self::read_dist_path(name, directory.span(), &dir_path, &dir_path)
    }
}

impl ToTokens for IncludeDist {
    fn to_tokens(&self, tokens: &mut proc_macro2::TokenStream) {
        // std::collections::HashMap
        let array_contents = self.files.iter().map(|file| {
            let FileWithPath { include, rel_path } = file;

            quote! {
                (#rel_path, #include)
            }
        });
        let name = &self.name;
        tokens.extend(quote! {
            pub const #name: &[(&'static str, &'static [u8])] = &[#(#array_contents),*];
        });
    }
}

#[proc_macro]
pub fn include_dist(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
    let incl_dist = parse_macro_input!(input as IncludeDist);
    quote! {#incl_dist}.into()
}
