use core::ops::Deref;
use std::{rc::Rc, sync::Arc};

use futures::{lock::Mutex, SinkExt, StreamExt};
use gloo::net::websocket::Message;
use scratchpad::{
    game::{self, BoardValues},
    ClientMessage, SendMove, ServerMessage,
};
use web_sys::{wasm_bindgen::JsCast, Element, HtmlDivElement, HtmlElement, HtmlTableCellElement};
use yew::prelude::*;

const IS_ON_CLASS: &str = "scratchpad-active-tile";

fn update_scratchpad(new_state: BoardValues) {
    let tiles_html = gloo::utils::document().get_elements_by_class_name("scratchpad-tile");
    let mut tiles = Vec::with_capacity(tiles_html.length() as _);
    for idx in 0..tiles_html.length() {
        let tile = tiles_html
            .get_with_index(idx)
            .unwrap()
            .dyn_into::<HtmlDivElement>()
            .unwrap();
        let row: u32 = tile.get_attribute("row").unwrap().parse().unwrap();
        let col: u32 = tile.get_attribute("col").unwrap().parse().unwrap();
        tiles.push((row, col, tile));
    }

    tiles.sort_by_key(|(row, col, _)| row * game::COLS as u32 + col);

    for (element, new_value) in tiles
        .into_iter()
        .map(|t| t.2)
        .zip(new_state.into_iter().flat_map(|row| row.into_iter()))
    {
        let currently = element.class_name().contains(IS_ON_CLASS);
        match (currently, new_value) {
            (true, false) => {
                element.set_class_name(element.class_name().replace(IS_ON_CLASS, "").as_str())
            }
            (false, true) => {
                element.set_class_name(format!("{} {IS_ON_CLASS}", element.class_name()).as_str())
            }
            _ => continue,
        }
    }
}

#[function_component]
pub fn ScratchPad() -> Html {
    //     let host = gloo::utils::window()
    //         .location()
    //         .host()
    //         .expect("No host found in current window");
    // let ws_url = format!("ws://{host}/scratchpad").replace(":8080", ":3031");
    let ws_url = String::from("wss://jigglypuff.club/scratchpad");

    let send = {
        let ws = gloo::net::websocket::futures::WebSocket::open(&ws_url).unwrap();
        let (send, mut recv) = ws.split();
        let send = Rc::new(Mutex::new(send));
        yew::platform::spawn_local(async move {
            loop {
                let msg = match recv.next().await {
                    Some(Ok(msg)) => msg,
                    Some(Err(err)) => {
                        log::error!("scratchpad socket error: {err}");
                        return;
                    }
                    None => {
                        log::error!("scratchpad socket closed");
                        return;
                    }
                };
                let bytes = match msg {
                    Message::Text(_) => {
                        log::error!("did not expect text message; discarding.");
                        continue;
                    }
                    Message::Bytes(bytes) => bytes,
                };
                if bytes.is_empty() {
                    log::warn!("empty message; discarding.");
                    continue;
                }

                let msg_type: scratchpad::ServerMessageType = match bytes[0].try_into() {
                    Ok(t) => t,
                    Err(err) => {
                        log::error!("invalid message: {err} ({:#04X})", bytes[0]);
                        continue;
                    }
                };
                let ServerMessage::StateUpdate(msg) = match msg_type.try_parse_message(&bytes[1..])
                {
                    Ok(msg) => msg,
                    Err(err) => {
                        log::error!("{err}");
                        continue; // im tired idc
                    }
                };
                update_scratchpad(msg);
            }
        });

        let send_cl = send.clone();
        yew::platform::spawn_local(async move {
            let board_req: Vec<u8> = ClientMessage::GetBoard.into();
            if let Err(err) = send_cl.lock().await.send(Message::Bytes(board_req)).await {
                log::error!("sending req: {err}");
            }
        });
        send
    };

    let grid_tile_click = || {
        let send_click = send.clone();
        // let state = ws_state.clone();
        move |event: MouseEvent| {
            // let (x, y) = state.deref();
            let target = event
                .target()
                .and_then(|t| t.dyn_into::<Element>().ok())
                .expect("grid_tile_click click target is not an element??");

            let row: u32 = target
                .get_attribute("row")
                .expect("no row attribute on grid_tile_click")
                .parse()
                .expect("row is not u32");

            let col: u32 = target
                .get_attribute("col")
                .expect("no col attribute on grid_tile_click")
                .parse()
                .expect("col is not u32");

            let value = match target.class_name().contains("scratchpad-active-tile") {
                true => {
                    target.set_class_name(target.class_name().replace(IS_ON_CLASS, "").as_str());
                    true
                }
                false => {
                    target
                        .set_class_name(format!("{} {IS_ON_CLASS}", target.class_name()).as_str());
                    false
                }
            };

            // flip cause it's setting now
            let value = !value;
            let send_click_clone = send_click.clone();
            yew::platform::spawn_local(async move {
                let bytes: Vec<u8> = ClientMessage::SendMove(SendMove { row, col, value }).into();
                if let Err(err) = send_click_clone
                    .lock()
                    .await
                    .send(Message::Bytes(bytes))
                    .await
                {
                    log::error!("Sending tile update: {err}");
                }
            });
        }
    };

    let rows = (0u32..scratchpad::game::ROWS as u32)
        .map(|row| (row, (0u32..scratchpad::game::COLS as u32)))
        .map(|(row_idx, cols)| {
			let columns = cols.map(move |col_idx| {
                html! {
                    <div class="scratchpad-tile" row={row_idx.to_string()} col={col_idx.to_string()} onclick={grid_tile_click()}>
						// <span />
					</div>
                }
            });
			html! {
				<>
					{for columns}
				</>
			}
        });

    html! {
        <div>
            <div class="scratchpad-grid">
                {for rows}
            </div>
        </div>
    }
}
