use game::{GameError, TileType};
use thiserror::Error;

pub mod game;

#[derive(Debug, Error, Clone, Copy)]
pub enum MessageError {
    #[error("invalid message type")]
    InvalidMessageType,
    #[error("message too short")]
    MessageTooShort,
    #[error("invalid tile value")]
    InvalidTileValue,
    #[error("{0}")]
    GameError(#[from] GameError),
}

#[repr(u8)]
#[derive(Debug, Clone, Copy, PartialEq)]
pub enum ClientMessageType {
    GetBoard = 0x10,
    SendMove = 0x20,
}

impl TryFrom<u8> for ClientMessageType {
    type Error = MessageError;

    fn try_from(value: u8) -> Result<Self, Self::Error> {
        Ok(match value {
            0x10 => ClientMessageType::GetBoard,
            0x20 => ClientMessageType::SendMove,
            _ => return Err(MessageError::InvalidMessageType),
        })
    }
}

#[derive(Debug)]
pub struct SendMove {
    pub row: u32,
    pub col: u32,
    pub value: TileType,
}

impl TryFrom<&[u8]> for SendMove {
    type Error = MessageError;

    fn try_from(value: &[u8]) -> Result<Self, Self::Error> {
        if value.len() < 9 {
            return Err(MessageError::InvalidMessageType);
        }
        let row = u32::from_be_bytes(value[..4].try_into().unwrap());
        let col = u32::from_be_bytes(value[4..8].try_into().unwrap());
        let value = match value[8] {
            0 => false,
            1 => true,
            _ => return Err(MessageError::InvalidTileValue),
        };
        Ok(Self { row, col, value })
    }
}

impl SendMove {
    fn write_to_vec(&self, out: &mut Vec<u8>) {
        out.extend_from_slice(&self.row.to_be_bytes());
        out.extend_from_slice(&self.col.to_be_bytes());
        out.push(self.value as _);
    }
}

#[derive(Debug)]
pub enum ClientMessage {
    GetBoard,
    SendMove(SendMove),
}

impl From<ClientMessage> for Vec<u8> {
    fn from(value: ClientMessage) -> Self {
        let mut out = Vec::new();
        match value {
            ClientMessage::GetBoard => out.push(ClientMessageType::GetBoard as u8),
            ClientMessage::SendMove(send_move) => {
                out.push(ClientMessageType::SendMove as u8);
                send_move.write_to_vec(&mut out);
            }
        }

        out
    }
}

impl ClientMessageType {
    pub fn try_parse_message(&self, bytes: &[u8]) -> Result<ClientMessage, MessageError> {
        match self {
            ClientMessageType::GetBoard => Ok(ClientMessage::GetBoard),
            ClientMessageType::SendMove => Ok(ClientMessage::SendMove(bytes.try_into()?)),
        }
    }
}

#[repr(u8)]
#[derive(Debug, Clone, Copy, PartialEq)]
pub enum ServerMessageType {
    StateUpdate = 0x30,
}

impl TryFrom<u8> for ServerMessageType {
    type Error = MessageError;

    fn try_from(value: u8) -> Result<Self, Self::Error> {
        Ok(match value {
            0x30 => ServerMessageType::StateUpdate,
            _ => return Err(MessageError::InvalidMessageType),
        })
    }
}
#[derive(Debug)]
pub enum ServerMessage {
    StateUpdate(game::BoardValues),
}

impl ServerMessageType {
    pub fn try_parse_message(&self, bytes: &[u8]) -> Result<ServerMessage, MessageError> {
        Ok(match self {
            ServerMessageType::StateUpdate => {
                ServerMessage::StateUpdate(game::board_values_from_bytes(bytes)?)
            }
        })
    }
}

impl From<ServerMessage> for Vec<u8> {
    fn from(value: ServerMessage) -> Self {
        match value {
            ServerMessage::StateUpdate(state) => [ServerMessageType::StateUpdate as u8]
                .into_iter()
                .chain(
                    state
                        .into_iter()
                        .flat_map(|row| row.into_iter())
                        .map(|tile| tile as u8),
                )
                .collect(),
        }
    }
}
